"""
Script to compare performance of Vanilla SGD, FederatedAvergaging, and LASG-WK2 on
the Nietzsche corpus.
Character-LSTM implementation follows this example at
https://keras.io/examples/generative/lstm_character_level_text_generation/
"""

import io
import numpy as np
import pandas as pd
from tensorflow import keras
from tensorflow.keras import layers
# from sklearn.model_selection import train_test_split
from tensorflow.keras.layers import LSTM, Bidirectional
import runner
import lasg
import federated_avg
import vanilla_sgd
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, LSTM, Dropout
from tensorflow.keras.layers import LSTM, Dense, Activation
from Attention_utils import Attention
from tensorflow.keras.models import Model
from tensorflow.python.keras.metrics import RootMeanSquaredError
from tensorflow.keras.layers import Input
from tensorflow.keras.optimizers import Adam
import argparse
import os

def run():
    """
    Execute the script
    """
    ############################################################################
    # SETUP
    ############################################################################

    X_train = pd.read_csv('./datasets/9_data2的副本.csv', header=None)
    X_train = np.array(X_train).astype('float32')
    X_test = pd.read_csv('./datasets/9_test4的副本.csv', header=None)
    X_test = np.array(X_test).astype('float32')
    y_train = pd.read_csv('./datasets/train_label3的副本.csv', header=None)
    y_train = np.array(y_train).astype('float32')
    y_test = pd.read_csv('./datasets/test_label3的副本.csv', header=None)
    y_test = np.array(y_test).astype('float32')
    y_train = np.squeeze(y_train)
    y_test = np.squeeze(y_test)

    parser = argparse.ArgumentParser()
    parser.add_argument('--model_name', type=str, default='BiLSTM_Attn')
    parser.add_argument('--epochs', type=int, default=600, help="Epochs")
    parser.add_argument('--batch_size', type=int, default=512, help="Batch Size")
    parser.add_argument('--learning_rate', type=float, default=0.001, help="Learning rate")
    parser.add_argument('--sequence_length', type=int, default=10, help="sequence length")
    parser.add_argument('--dataset', type=str, default='load', help="Training set")
    parser.add_argument('--path', type=str, default='./results/')
    args = parser.parse_args([])
    if not os.path.exists(args.path):
        os.mkdir(args.path)

    def convertSeriesToMatrix(vectorSeries, sequence_length):
        """
        滑动时间窗口处理
        """
        matrix = []
        for i in range(len(vectorSeries) - sequence_length + 1):
            matrix.append(vectorSeries[i: i + sequence_length])
        return matrix

    #Define the model, loss, and optimizer
    def model_constructor(hparams):

        input_len = args.sequence_length - 1

        model_input = Input(shape=(input_len, 1))
        x = Bidirectional(LSTM(50, return_sequences=True))(model_input)
        x = Bidirectional(LSTM(100, return_sequences=True))(x)
        x = Attention(32)(x)
        output = Dense(units=1)(x)
        model = Model(inputs=model_input, outputs=output)  # optimizer=Adam(learning_rate=learning_rate, clipnorm=1.0)
        model.compile(loss='mean_squared_error', optimizer=Adam(learning_rate=0.001, clipnorm=0.5),
                      metrics=['mae', RootMeanSquaredError()])
        model.summary()
        print('inputs')

        # plot_model(model, show_shapes=True, to_file=args.path + args.model_name + '_structure_img.png')
        print('-------- Dataset: {} --------'.format(args.dataset))
        print('-------- Training Loop for model {}-------- '.format(args.model_name))
        history = model.fit(X_train, y_train, batch_size=args.batch_size, epochs=args.epochs, validation_split=0.05,
                            verbose=1)
        # model.save(args.model_name + '_model.h5')  # 保存模型

        return model
        # model = Sequential()
        # model.add(LSTM(units=24, input_dim=1, kernel_initializer='lecun_uniform', return_sequences=False))
        # model.add(Dense(units=1))  # optimizers.Adam(learning_rate=learning_rate, clipnorm=1.0)


        return model
    ############################################################################
    # RUN EXPERIMENTS
    ############################################################################

    #Run each experiment until this target test accuracy is reached
    target_test_accuracy = 0.17

    # vanilla_sgd_hparams = vanilla_sgd.vanilla_sgd_hparams(eta=1.5, batch_size=32, epochs=100, evaluation_interval=5,
    #                                                       target_val_accuracy=target_test_accuracy)
    #
    #The hyperparameter variables for the federated learning algorithms are set to align with their respective papers.
    #This can cause confusion: for example the hparam 'K' means # of workers for FederatedAveraging and # of iterations for LASG.
    #See the constructor docstrings for a reference on what each hyperparameter does for each algorithm.

    fedavg_hparams = federated_avg.fedavg_hparams(K=5, C=0.1, E=1, B=10, eta=1.5, MAX_T=10000, evaluation_interval=2,
                                                  target_val_accuracy=target_test_accuracy)
    #
    # lasg_wk2_hparams = lasg.lasg_wk2_hparams(M=10, K=10000, D=50, c=0.1, c_range=10, eta=1.5, B=100, evaluation_interval=5,
    #                                          target_val_accuracy=target_test_accuracy,
    #                                          print_lasg_wk2_condition=True)
    #
    #Run with the training set distributed I.I.D
    print()
    print("Running Experiments with training set distributed I.I.D")
    print()

    # vanilla_sgd_hparams.iid = True
    fedavg_hparams.iid = True
    # lasg_wk2_hparams.iid = True

    iid_vanilla_sgd_log, iid_fedavg_log, iid_lasg_wk2_log = runner.run_experiments(
                    X_train, y_train, X_test, y_test, model_constructor,
                     fedavg_hparams,  seed=100)

    #Run with the training set distributed non-I.I.D
    #We use the the technique of McMahan, et al. to split the data among workers in a
    #"pathologically" non-IID manner where each worker may only have a few labels represented
    #in its local data. Additionally, we also randomize the volume of data each worker receives.
    print()
    print("Running Experiments with training set distributed non-I.I.D")
    print()

    # vanilla_sgd_hparams.iid = False
    # fedavg_hparams.iid = False
    # lasg_wk2_hparams.iid = False
    #
    # non_iid_vanilla_sgd_log, non_iid_fedavg_log, non_iid_lasg_wk2_log = runner.run_experiments(
    #                 X_train, y_train, X_test, y_test, model_constructor,
    #                 vanilla_sgd_hparams, fedavg_hparams, lasg_wk2_hparams, seed=100)
    #
    # print()
    # print("Done!")
    # print()
    #
    ############################################################################
    # Plot results
    ############################################################################
    runner.plot_results(iid_vanilla_sgd_log, iid_fedavg_log, iid_lasg_wk2_log,
                        target_test_accuracy, iid=True, dataset="Nietzsche", modeltype="LSTM", mode="save")

    # #runner.plot_results(non_iid_vanilla_sgd_log, non_iid_fedavg_log, non_iid_lasg_wk2_log,
    #                     target_test_accuracy, iid=False, dataset="Nietzsche", modeltype="LSTM", mode="save")
    #
if __name__ == "__main__":
    run()